/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.aws.lambda;

import com.amazonaws.services.lambda.runtime.Client;
import com.amazonaws.services.lambda.runtime.ClientContext;
import com.amazonaws.services.lambda.runtime.Context;
import com.amazonaws.services.lambda.runtime.RequestStreamHandler;
import com.cloudera.cdp.shaded.org.apache.commons.lang3.StringUtils;
import com.cloudera.naaf.StandardStatelessNiFiFunction;
import com.cloudera.naaf.StatelessFunctionConfiguration;
import com.cloudera.naaf.StatelessNiFiFunction;
import com.cloudera.naaf.metering.CloudPlatform;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import static com.cloudera.naaf.StatelessNiFiFunction.DEFAULT_PARAMETER_CONTEXT_VAR;
import static com.cloudera.naaf.StatelessNiFiFunction.PARAM_CONTEXT_MAPPING_PREFIX;

public class StatelessNiFiFunctionHandler implements RequestStreamHandler {
    private static final Logger logger = LoggerFactory.getLogger(StatelessNiFiFunctionHandler.class);

    // Environment Variables
    private static final String FUNCTION_NAME_ENV_VAR = "AWS_LAMBDA_FUNCTION_NAME";
    private static final String FUNCTION_VERSION_ENV_VAR = "AWS_LAMBDA_FUNCTION_VERSION";
    private static final String STORAGE_EXTENSIONS_DIRECTORY_ENV_VAR = "STORAGE_EXTENSIONS_DIRECTORY";
    private static final String STORAGE_RESOURCES_DIRECTORY_ENV_VAR = "STORAGE_RESOURCES_DIRECTORY";
    private static final String STORAGE_BUCKET_ENV_VAR = "STORAGE_BUCKET";
    private static final String AWS_REGION_ENV_VAR = "AWS_REGION";

    // Defaults
    private static final String DEFAULT_STORAGE_EXTENSIONS_DIRECTORY = "extensions";
    private static final String DEFAULT_STORAGE_RESOURCES_DIRECTORY = "resources";
    private static final String RESOURCE_DOWNLOADER_TYPE = "com.cloudera.naaf.aws.lambda.resources.AwsS3ResourceDownloader";
    private static final String STATE_PROVIDER_TYPE = "com.cloudera.naaf.aws.lambda.DynamoDBStatelessNiFiStateProvider";

    private static final File RESOURCES_DIRECTORY = new File(StatelessNiFiFunction.TEMP_DIR, "resources");
    private static final File DEFAULT_EXTENSION_DIRECTORY = new File(StatelessNiFiFunction.TEMP_DIR, "extensions");

    private static final String AWS_SECRETS_MANAGER_PROPERTY_PREFIX = "nifi.stateless.parameter.provider.AWSSecretsManager-%d.";
    private static final String ATTRIBUTE_PREFIX = "aws.lambda.context.";

    private static final String AWS_CLOUD_PROVIDER = "AWS";
    private static final String POM_PROPERTIES = "pom.properties";
    private static final String NIFI_VERSION = "nifi.version";

    private final StatelessNiFiFunction statelessNiFiFunction;
    private final PropertiesReader propertiesReader;

    public StatelessNiFiFunctionHandler() throws IOException {
        final String functionName = System.getenv(FUNCTION_NAME_ENV_VAR);
        final String functionVersion = System.getenv(FUNCTION_VERSION_ENV_VAR);

        final File resourcesDirectory = RESOURCES_DIRECTORY;
        if (!resourcesDirectory.exists() && !resourcesDirectory.mkdirs()) {
            throw new IOException("Resources directory " + resourcesDirectory.getAbsolutePath() + " does not exist and cannot be created");
        }

        String resourcesBucketDirectory = null;
        String extensionsBucketDirectory = null;
        final String storageBucket = System.getenv(STORAGE_BUCKET_ENV_VAR);

        if (storageBucket != null) {
            resourcesBucketDirectory = Optional.ofNullable(System.getenv(STORAGE_RESOURCES_DIRECTORY_ENV_VAR))
                    .orElse(DEFAULT_STORAGE_RESOURCES_DIRECTORY);
            extensionsBucketDirectory = Optional.ofNullable(System.getenv(STORAGE_EXTENSIONS_DIRECTORY_ENV_VAR))
                    .orElse(DEFAULT_STORAGE_EXTENSIONS_DIRECTORY);
        }

        final StatelessFunctionConfiguration configurationActions = new StatelessFunctionConfiguration.Builder()
                .region(System.getenv(AWS_REGION_ENV_VAR))
                .cloudPlatform(CloudPlatform.AWS_LAMBDA)
                .flowDefinitionConfigAction(this::enableAwsSecretsManager)
                .resourcesDirectory(resourcesDirectory)
                .resourcesSourceDirectory(resourcesBucketDirectory)
                .extensionsSourceDirectory(extensionsBucketDirectory)
                .defaultExtensionDirectory(DEFAULT_EXTENSION_DIRECTORY)
                .resourceDownloaderType(RESOURCE_DOWNLOADER_TYPE)
                .stateProviderType(STATE_PROVIDER_TYPE)
                .systemClassLoader(this.getClass().getClassLoader())
                .build();
        propertiesReader = new PropertiesReader(POM_PROPERTIES);
        statelessNiFiFunction = new StandardStatelessNiFiFunction(functionName, functionVersion, configurationActions);
    }

    @Override
    public void handleRequest(final InputStream inputStream, final OutputStream outputStream, final Context context) throws IOException {
        final String instanceId = context.getInvokedFunctionArn();
        statelessNiFiFunction.trigger(instanceId, inputStream, getAttributes(context), outputStream);
    }

    private Map<String, String> getAttributes(final Context context) {
        final Map<String, String> attributes = new HashMap<>();
        attributes.put(StatelessNiFiFunction.CLOUD_PROVIDER_ATTRIBUTE, AWS_CLOUD_PROVIDER);
        attributes.put(ATTRIBUTE_PREFIX + "awsRequestId", context.getAwsRequestId());
        attributes.put(ATTRIBUTE_PREFIX + "functionName", context.getFunctionName());
        attributes.put(ATTRIBUTE_PREFIX + "functionVersion", context.getFunctionVersion());
        attributes.put(ATTRIBUTE_PREFIX + "invokedFunctionArn", context.getInvokedFunctionArn());
        attributes.put(ATTRIBUTE_PREFIX + "logGroupName", context.getLogGroupName());
        attributes.put(ATTRIBUTE_PREFIX + "logStreamName", context.getLogStreamName());
        attributes.put(ATTRIBUTE_PREFIX + "memoryLimitInMB", String.valueOf(context.getMemoryLimitInMB()));
        attributes.put(ATTRIBUTE_PREFIX + "remainingTimeInMillis", String.valueOf(context.getRemainingTimeInMillis()));
        final ClientContext clientContext = context.getClientContext();
        if (clientContext != null) {
            final Client client = clientContext.getClient();
            if (client != null) {
                attributes.put(ATTRIBUTE_PREFIX + "client.installationId", client.getInstallationId());
                attributes.put(ATTRIBUTE_PREFIX + "client.appPackageName", client.getAppPackageName());
                attributes.put(ATTRIBUTE_PREFIX + "client.appTitle", client.getAppTitle());
                attributes.put(ATTRIBUTE_PREFIX + "client.appVersionCode", client.getAppVersionCode());
                attributes.put(ATTRIBUTE_PREFIX + "client.appVersionName", client.getAppVersionName());
            }
            if (clientContext.getEnvironment() != null) {
                clientContext.getEnvironment().forEach((key, value) -> attributes.put(ATTRIBUTE_PREFIX + "client.environment." + key, value));
            }
        }
        return attributes;
    }

    @Override
    protected void finalize() {
        statelessNiFiFunction.cleanUp();
    }

    private void enableAwsSecretsManager(final Map<String, String> properties) {
        enableAwsSecretsManager(properties, System.getenv(), propertiesReader);
    }

    static void enableAwsSecretsManager(final Map<String, String> properties, final Map<String, String> environmentVariables,
                                        final PropertiesReader propertiesReader) {

        final String defaultParamContextName = System.getenv(DEFAULT_PARAMETER_CONTEXT_VAR);

        final Map<String, Set<String>> secretNameToContextNameMap = new HashMap<>();

        // If an Environment Variable exists that starts with PARAM_CONTEXT_ then map the rest of the env variable name to a secret name.
        // For example, if ENV variable PARAM_CONTEXT_ABC = xyz, then the values for the "ABC" parameter context should be pulled from the AWS Secret named "xyz".
        environmentVariables.forEach((name, secretNameString) -> {
            if (name.startsWith(PARAM_CONTEXT_MAPPING_PREFIX) && name.length() > PARAM_CONTEXT_MAPPING_PREFIX.length()) {
                final String paramContextName = name.substring(PARAM_CONTEXT_MAPPING_PREFIX.length());
                final String[] secretNames = secretNameString.split("\\s*,\\s*");
                for (final String secretName : secretNames) {
                    secretNameToContextNameMap.computeIfAbsent(secretName.trim(), k -> new HashSet<>()).add(paramContextName);
                }
            }
        });

        final AtomicInteger i = new AtomicInteger(0);
        secretNameToContextNameMap.forEach((secretName, paramContextNames) -> {
            final String propertiesPrefix = String.format(AWS_SECRETS_MANAGER_PROPERTY_PREFIX, i.getAndIncrement());
            addSecretsManagerProperties(propertiesPrefix, properties, secretName, defaultParamContextName, propertiesReader);
            paramContextNames.forEach(paramContextName -> properties.put(propertiesPrefix + "properties." + paramContextName, secretName));
            final String key = StringUtils.substringAfterLast(StringUtils.substringBeforeLast(propertiesPrefix, "."), ".");
            logger.debug("Added [{}}], mapping Secret [{}] to Parameter Contexts {}", key, secretName, paramContextNames);
        });
        // If there are no explicit mappings, add a provider for the default parameter context secret
        if (i.get() == 0) {
            final String propertiesPrefix = String.format(AWS_SECRETS_MANAGER_PROPERTY_PREFIX, i.getAndIncrement());
            addSecretsManagerProperties(propertiesPrefix, properties, "Default", defaultParamContextName, propertiesReader);
        }
    }

    private static void addSecretsManagerProperties(final String propertyPrefix, final Map<String, String> properties, final String secretName,
                                             final String defaultParamContextName, final PropertiesReader propertiesReader) {
        final String nifiVersion = propertiesReader.getProperty(NIFI_VERSION);
        properties.put(propertyPrefix + "name", String.format("AWS Secrets Manager Parameter Value Provider [%s]", secretName));
        properties.put(propertyPrefix + "bundle", "org.apache.nifi:nifi-aws-nar:" + nifiVersion);
        properties.put(propertyPrefix + "type", "org.apache.nifi.stateless.parameter.AwsSecretsManagerParameterValueProvider");

        if (defaultParamContextName != null && !defaultParamContextName.trim().isEmpty()) {
            properties.put(propertyPrefix + "properties.default-secret-name", defaultParamContextName);
        }
    }
}
