/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.aws.lambda;

import com.cloudera.naaf.StatelessNiFiStateProvider;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.auth.credentials.DefaultCredentialsProvider;
import software.amazon.awssdk.services.dynamodb.DynamoDbClient;
import software.amazon.awssdk.services.dynamodb.model.AttributeDefinition;
import software.amazon.awssdk.services.dynamodb.model.AttributeValue;
import software.amazon.awssdk.services.dynamodb.model.BillingMode;
import software.amazon.awssdk.services.dynamodb.model.CreateTableRequest;
import software.amazon.awssdk.services.dynamodb.model.DescribeTableRequest;
import software.amazon.awssdk.services.dynamodb.model.GetItemRequest;
import software.amazon.awssdk.services.dynamodb.model.GetItemResponse;
import software.amazon.awssdk.services.dynamodb.model.KeySchemaElement;
import software.amazon.awssdk.services.dynamodb.model.KeyType;
import software.amazon.awssdk.services.dynamodb.model.ProvisionedThroughput;
import software.amazon.awssdk.services.dynamodb.model.PutItemRequest;
import software.amazon.awssdk.services.dynamodb.model.ResourceNotFoundException;
import software.amazon.awssdk.services.dynamodb.model.ScalarAttributeType;
import software.amazon.awssdk.services.dynamodb.model.TableStatus;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

public class DynamoDBStatelessNiFiStateProvider implements StatelessNiFiStateProvider {
    private static final Logger logger = LoggerFactory.getLogger(DynamoDBStatelessNiFiStateProvider.class);

    private static final String DYNAMODB_STATE_TABLE_VAR = "DYNAMODB_STATE_TABLE";
    private static final String DYNAMODB_BILLING_MODE_VAR = "DYNAMODB_BILLING_MODE";
    private static final String DYNAMODB_WRITE_CAPACITY_UNITS_VAR = "DYNAMODB_WRITE_CAPACITY_UNITS";
    private static final String DYNAMODB_READ_CAPACITY_UNITS_VAR = "DYNAMODB_READ_CAPACITY_UNITS";

    private static final String DEFAULT_STATE_TABLE = "nifi_state";
    public static final String INSTANCE_ID = "instanceId";
    public static final String SCOPE = "scope";
    public static final String STATE = "state";

    private final BillingMode billingMode;
    private final Long readCapacityUnits;
    private final Long writeCapacityUnits;
    private final String tableName;

    private final DynamoDbClient dynamoDbClient;
    private final ObjectMapper objectMapper = new ObjectMapper();

    public DynamoDBStatelessNiFiStateProvider() {
        dynamoDbClient = DynamoDbClient.builder()
                .credentialsProvider(DefaultCredentialsProvider.builder().build())
                .build();

        tableName = Optional.ofNullable(System.getenv(DYNAMODB_STATE_TABLE_VAR)).orElse(DEFAULT_STATE_TABLE);

        final String billingModeValue = System.getenv(DYNAMODB_BILLING_MODE_VAR);
        if (billingModeValue == null) {
            logger.info("Defaulting DynamoDB Billing Mode to PAY_PER_REQUEST for State Provider");
            billingMode = BillingMode.PAY_PER_REQUEST;
        } else {
            billingMode = BillingMode.fromValue(billingModeValue);
        }

        if (billingMode == BillingMode.PROVISIONED) {
            readCapacityUnits = getCapacityUnits(DYNAMODB_READ_CAPACITY_UNITS_VAR);
            writeCapacityUnits = getCapacityUnits(DYNAMODB_WRITE_CAPACITY_UNITS_VAR);
        } else {
            readCapacityUnits = null;
            writeCapacityUnits = null;
        }

        TableStatus tableStatus = getTableStatus(tableName);
        if (tableStatus == null) {
            tableStatus = createTable(tableName);
        }
        tableStatus = handleTableStatus(tableName, tableStatus);
        if (tableStatus != TableStatus.ACTIVE) {
            logger.error("DynamoDB table {} was in invalid state {}, so NiFi's state cannot be stored or retrieved", tableName, tableStatus);
            throw new RuntimeException(String.format("DynamoDB table %s was in invalid state %s, so NiFi's state cannot be stored or retrieved", tableName, tableStatus));
        }
    }

    private Long getCapacityUnits(final String environmentVariableName) {
        final String value = System.getenv(environmentVariableName);
        if (value == null) {
            throw new IllegalArgumentException(environmentVariableName + " Environment Variable is required when " + DYNAMODB_BILLING_MODE_VAR + " is PROVISIONED");
        }
        return Long.parseLong(value);
    }

    @Override
    public void storeState(final String instanceId, final StateScope scope, final Map<String, String> state) {
        final Map<String, AttributeValue> attributes = new HashMap<>();
        final String stateJson;
        try {
            stateJson = objectMapper.writeValueAsString(state);
        } catch (final JsonProcessingException e) {
            throw new RuntimeException("Could not write state for instance " + instanceId, e);
        }
        attributes.put(INSTANCE_ID, AttributeValue.builder().s(instanceId).build());
        attributes.put(SCOPE, AttributeValue.builder().s(scope.name()).build());
        attributes.put(STATE, AttributeValue.builder().s(stateJson).build());
        final PutItemRequest request = PutItemRequest.builder().tableName(tableName).item(attributes).build();
        dynamoDbClient.putItem(request);
    }

    @Override
    public Map<String, String> retrieveState(final String instanceId, final StateScope scope) {
        final Map<String, AttributeValue> key = new HashMap<>();
        key.put(INSTANCE_ID, AttributeValue.builder().s(instanceId).build());
        key.put(SCOPE, AttributeValue.builder().s(scope.name()).build());
        final GetItemRequest request = GetItemRequest.builder().tableName(tableName).key(key).build();
        final GetItemResponse response = dynamoDbClient.getItem(request);
        final AttributeValue stateValue = response.item().get(STATE);
        if (stateValue == null) {
            logger.info("State attribute has not been stored for {}, scope {}", instanceId, scope);
            return null;
        }

        final String stateJson = stateValue.s();
        try {
            final Map<String, String> state = objectMapper.readValue(stateJson, Map.class);
            return state;
        } catch (final JsonProcessingException e) {
            throw new IllegalStateException("Failed to parse state for instance " + instanceId, e);
        }
    }

    private TableStatus createTable(final String instanceId) {
        final CreateTableRequest.Builder builder = CreateTableRequest.builder()
                .tableName(instanceId)
                .attributeDefinitions(AttributeDefinition.builder().attributeName(INSTANCE_ID).attributeType(ScalarAttributeType.S).build(),
                        AttributeDefinition.builder().attributeName(SCOPE).attributeType(ScalarAttributeType.S).build())
                .keySchema(KeySchemaElement.builder().attributeName(INSTANCE_ID).keyType(KeyType.HASH).build(),
                        KeySchemaElement.builder().attributeName(SCOPE).keyType(KeyType.RANGE).build())
                .billingMode(billingMode);
        if (billingMode == BillingMode.PROVISIONED) {
            builder.provisionedThroughput(ProvisionedThroughput.builder()
                    .readCapacityUnits(readCapacityUnits)
                    .writeCapacityUnits(writeCapacityUnits)
                    .build());
        }
        dynamoDbClient.createTable(builder.build());

        return getTableStatus(instanceId);
    }

    private TableStatus handleTableStatus(final String tableName, TableStatus tableStatus) {
        switch (tableStatus) {
            case CREATING:
            case UPDATING:
                while (TableStatus.ACTIVE != tableStatus) {
                    tableStatus = getTableStatus(tableName);
                    try {
                        Thread.sleep(1000L);
                    } catch (final InterruptedException e) {
                        throw new RuntimeException("Interrupted thread", e);
                    }
                }
                break;
        }
        return tableStatus;
    }

    private TableStatus getTableStatus(final String tableName) {
        final DescribeTableRequest request = DescribeTableRequest.builder().tableName(tableName).build();
        try {
            return dynamoDbClient.describeTable(request).table().tableStatus();
        } catch (final ResourceNotFoundException e) {
            logger.info("Table {} does not exist: creating it", tableName);
            return null;
        }
    }
}
