/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.aws.lambda.resources;

import com.cloudera.naaf.resources.AbstractResourceDownloader;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.auth.credentials.DefaultCredentialsProvider;
import software.amazon.awssdk.core.ResponseInputStream;
import software.amazon.awssdk.services.s3.S3Client;
import software.amazon.awssdk.services.s3.model.GetObjectRequest;
import software.amazon.awssdk.services.s3.model.GetObjectResponse;
import software.amazon.awssdk.services.s3.model.ListObjectsRequest;
import software.amazon.awssdk.services.s3.model.ListObjectsResponse;
import software.amazon.awssdk.services.s3.model.S3Object;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

public class AwsS3ResourceDownloader extends AbstractResourceDownloader {
    private static final Logger logger = LoggerFactory.getLogger(AwsS3ResourceDownloader.class);

    private static final String STORAGE_BUCKET_ENV_VAR = "STORAGE_BUCKET";

    @Override
    public void download(final String sourceDirectory, final File downloadDirectory) {
        final String bucketName = System.getenv(STORAGE_BUCKET_ENV_VAR);
        final String bucketPrefix = sourceDirectory.endsWith("/") ? sourceDirectory : sourceDirectory + "/";

        final S3Client s3 = S3Client.builder()
                .credentialsProvider(DefaultCredentialsProvider.builder().build())
                .build();

        ListObjectsRequest listObjectsRequest = ListObjectsRequest.builder()
                .bucket(bucketName)
                .prefix(bucketPrefix)
                .build();
        ListObjectsResponse response = s3.listObjects(listObjectsRequest);

        int downloadCount = 0;
        List<S3Object> objects = response.contents();

        boolean isTruncated = false;
        do {
            for (final S3Object object : objects) {
                isTruncated = response.isTruncated();
                final String key = object.key();
                final Long size = object.size();
                final String filenameWithinSourceDir = substringAfter(key, bucketPrefix);
                final Path filePath = getDownloadPath(filenameWithinSourceDir, downloadDirectory);
                if (filePath == null) {
                    continue;
                }

                logger.info("Downloading {} to {}", key, filePath);
                final GetObjectRequest getObjectRequest = GetObjectRequest.builder().bucket(bucketName).key(key).build();
                try (final ResponseInputStream<GetObjectResponse> responseInputStream = s3.getObject(getObjectRequest)) {
                    Files.copy(responseInputStream, filePath);
                } catch (final IOException e) {
                    logger.error("Failed to download {}", key, e);
                    throw new RuntimeException(String.format("Failed to download %s to %s", key, filePath));
                }
                logger.debug("Downloaded {} ({} bytes)", filePath, size);
                downloadCount++;
            }

            if (isTruncated) {
                final String nextMarker = response.nextMarker();
                listObjectsRequest = ListObjectsRequest.builder()
                        .bucket(bucketName)
                        .prefix(bucketPrefix)
                        .marker(nextMarker)
                        .build();
                response = s3.listObjects(listObjectsRequest);
                objects = response.contents();
            }
        } while (isTruncated);

        if (downloadCount == 0) {
            logger.debug("Found no objects in {}/{}", bucketName, bucketPrefix);
        }
    }
}
