/*
 * (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.azure.functions;

import com.cloudera.naaf.StatelessNiFiFunction;
import com.cloudera.naaf.StatelessNiFiUtil;
import com.microsoft.azure.functions.ExecutionContext;
import com.microsoft.azure.functions.HttpRequestMessage;
import com.microsoft.azure.functions.HttpResponseMessage;
import com.microsoft.azure.functions.HttpStatus;
import org.apache.nifi.flowfile.FlowFile;
import org.apache.nifi.stateless.flow.TriggerResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicReference;
import java.util.regex.Pattern;

/**
 * Azure Function with HTTP Trigger.
 */
public class AbstractStatelessNiFiHttpTriggerFunction {
    private static final Logger logger = LoggerFactory.getLogger(AbstractStatelessNiFiHttpTriggerFunction.class);

    private static final String ATTRIBUTE_PREFIX = "azure.http.header.";
    private static final String MIME_TYPE = "mime.type";

    // Environment variables
    private static final String HEADER_ATTRIBUTE_PATTERN_ENV_VAR = "HEADER_ATTRIBUTE_PATTERN";
    private static final String HTTP_STATUS_CODE_ATTRIBUTE_ENV_VAR = "HTTP_STATUS_CODE_ATTRIBUTE";

    private final Optional<Pattern> headerAttributePattern;

    private static final AtomicReference<StatelessNiFiAzureFunctionHolder> functionHolderReference = new AtomicReference<>();

    public AbstractStatelessNiFiHttpTriggerFunction() {
        final String headerAttributePatternVariable = System.getenv(HEADER_ATTRIBUTE_PATTERN_ENV_VAR);

        if (headerAttributePatternVariable != null) {
            headerAttributePattern = Optional.of(Pattern.compile(headerAttributePatternVariable));
        } else {
            headerAttributePattern = Optional.empty();
        }
    }

    private StatelessNiFiAzureFunctionHolder getFunctionHolder() {
        StatelessNiFiAzureFunctionHolder functionHolder = functionHolderReference.get();
        if (functionHolder == null) {
            logger.debug("Function has not been initialized yet, initializing now...");
            try {
                functionHolder = new StatelessNiFiAzureFunctionHolder();
            } catch (final IOException e) {
                logger.error("Error initializing function", e);
                throw new RuntimeException(e);
            }
            functionHolderReference.set(functionHolder);
            logger.debug("Successfully initialized Trigger Function");

            return functionHolder;
        }

        logger.debug("Function is already initialized");
        return functionHolder;
    }

    protected HttpResponseMessage runFunction(final HttpRequestMessage<?> request, final InputStream inputStream, final ExecutionContext executionContext) {
        HttpResponseMessage responseMessage = null;
        try {
            final StatelessNiFiAzureFunctionHolder functionHolder = getFunctionHolder();
            final StatelessNiFiFunction statelessNiFiFunction = functionHolder.getStatelessNiFiFunction();
            final Map<String, String> inputAttributes = StatelessNiFiAzureFunctionHolder.getAttributes(executionContext);
            inputAttributes.putAll(getAttributes(request));
            final String instanceId = functionHolder.getFunctionName();
            final TriggerResult triggerResult = statelessNiFiFunction.trigger(instanceId, inputStream, inputAttributes);
            logger.debug("Received {} trigger result{}", triggerResult.isSuccessful() ? "successful" : "failure",
                    triggerResult.getFailureCause().isPresent() ? " with failure cause " + triggerResult.getFailureCause().get().getMessage() : "");

            if (triggerResult.isSuccessful()) {
                final FlowFile result = statelessNiFiFunction.getOutputFlowfile(triggerResult);
                final Map<String, String> outputAttributes;
                if (result == null) {
                    logger.debug("Received no Flowfile output from dataflow");
                    outputAttributes = null;
                } else {
                    outputAttributes = result.getAttributes();
                }

                if (outputAttributes == null) {
                    responseMessage = request.createResponseBuilder(HttpStatus.OK).body("DataFlow completed successfully but had no output").build();
                } else {
                    final ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                    StatelessNiFiUtil.copyOutput(result, triggerResult, outputStream);
                    responseMessage = buildHttpResponse(request, outputStream.toByteArray(), outputAttributes);
                }
            }

            // If there was a failure, this will throw an exception
            StatelessNiFiUtil.handleTriggerResult(triggerResult);
        } catch (final Throwable t) {
            logger.error("DataFlow failed to execute", t);
            responseMessage = request.createResponseBuilder(HttpStatus.INTERNAL_SERVER_ERROR)
                    .body("DataFlow failure: " + t.getMessage())
                    .build();
        }
        return responseMessage;
    }

    private Map<String, String> getAttributes(final HttpRequestMessage<?> httpRequestMessage) {
        final Map<String, String> attributes = new HashMap<>();
        httpRequestMessage.getHeaders().forEach( (key, value) -> {
            attributes.put(StatelessNiFiFunction.CLOUD_PROVIDER_ATTRIBUTE, StatelessNiFiAzureFunctionHolder.AZURE_CLOUD_PROVIDER);
            attributes.put(ATTRIBUTE_PREFIX + key, value);
            if ("Content-Type".equalsIgnoreCase(key)) {
                attributes.put(MIME_TYPE, value);
            }
        });
        attributes.put("azure.http.method", httpRequestMessage.getHttpMethod().name());
        return attributes;
    }

    private HttpResponseMessage buildHttpResponse(final HttpRequestMessage<?> request,
                                   final byte[] responseBody, final Map<String, String> outputAttributes) {
        HttpStatus status = HttpStatus.OK;
        final HttpResponseMessage.Builder builder = request.createResponseBuilder(status);
        if (outputAttributes != null) {
            if (outputAttributes.containsKey(MIME_TYPE)) {
                builder.header("Content-Type", outputAttributes.get(MIME_TYPE));
            }
            final String httpStatusCodeAttribute = System.getenv(HTTP_STATUS_CODE_ATTRIBUTE_ENV_VAR);
            if (httpStatusCodeAttribute != null) {
                final String statusCodeValue = outputAttributes.get(httpStatusCodeAttribute);
                if (statusCodeValue != null) {
                    try {
                        final int parsedStatusCode = Integer.parseInt(statusCodeValue.trim());
                        status = HttpStatus.valueOf(parsedStatusCode);
                    } catch (final NumberFormatException e) {
                        final String errorMessage = String.format("Could not parse HTTP status code attribute [%s] value [%s]",
                                httpStatusCodeAttribute, statusCodeValue);
                        logger.error(errorMessage, e);
                        throw new RuntimeException(errorMessage, e);
                    }
                }
            }

            headerAttributePattern.ifPresent(attributePattern -> {
                outputAttributes.entrySet().stream()
                        .filter(entry -> attributePattern.matcher(entry.getKey()).matches())
                        .forEach(entry -> builder.header(entry.getKey(), entry.getValue()));
            });

        }
        builder.body(responseBody);
        builder.status(status);
        return builder.build();
    }
}
