/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.azure.functions;

import com.azure.cosmos.ConsistencyLevel;
import com.azure.cosmos.CosmosClient;
import com.azure.cosmos.CosmosClientBuilder;
import com.azure.cosmos.CosmosContainer;
import com.azure.cosmos.CosmosDatabase;
import com.azure.cosmos.implementation.NotFoundException;
import com.azure.cosmos.models.CosmosItemRequestOptions;
import com.azure.cosmos.models.CosmosItemResponse;
import com.azure.cosmos.models.PartitionKey;
import com.azure.identity.DefaultAzureCredentialBuilder;
import com.cloudera.naaf.StatelessNiFiStateProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;

public class CosmosDBStatelessNiFiStateProvider implements StatelessNiFiStateProvider {
    private static final Logger logger = LoggerFactory.getLogger(CosmosDBStatelessNiFiStateProvider.class);

    private static final String COSMOSDB_STATE_CONTAINER_VAR = "COSMOSDB_STATE_CONTAINER";
    private static final String COSMOSDB_STATE_DATABASE_VAR = "COSMOSDB_STATE_DATABASE";
    private static final String COSMOSDB_ENDPOINT_VAR = "COSMOSDB_ENDPOINT";

    private static final String DEFAULT_STATE_DATABASE = "nifi_state";
    private static final String DEFAULT_STATE_CONTAINER = "nifi_state";

    private final CosmosClient cosmosClient;
    private final String containerName;
    private final String databaseName;
    private final String endpoint;
    private final CosmosDatabase database;
    private CosmosContainer container;

    public CosmosDBStatelessNiFiStateProvider() {
        databaseName = Optional.ofNullable(System.getenv(COSMOSDB_STATE_DATABASE_VAR)).orElse(DEFAULT_STATE_DATABASE);
        containerName = Optional.ofNullable(System.getenv(COSMOSDB_STATE_CONTAINER_VAR)).orElse(DEFAULT_STATE_CONTAINER);
        endpoint = System.getenv(COSMOSDB_ENDPOINT_VAR);
        if (endpoint == null) {
            throw new IllegalArgumentException(COSMOSDB_ENDPOINT_VAR + " is required in order to use the CosmosDB state provider");
        }

        cosmosClient = new CosmosClientBuilder()
                .credential(new DefaultAzureCredentialBuilder().build())
                .endpoint(endpoint)
                .consistencyLevel(ConsistencyLevel.SESSION)
                .contentResponseOnWriteEnabled(true)
                .buildClient();

        database = cosmosClient.getDatabase(databaseName);
        container = database.getContainer(containerName);
    }

    @Override
    public void storeState(final String instanceId, final StateScope scope, final Map<String, String> state) {
        final String documentId = getDocumentId(instanceId, scope);
        final StateDocument stateDocument = new StateDocument(documentId, scope.name(), state);
        container.upsertItem(stateDocument, new PartitionKey(documentId), new CosmosItemRequestOptions());
    }

    private String getDocumentId(final String instanceId, final StateScope scope) {
        return UUID.nameUUIDFromBytes((instanceId + scope).getBytes(StandardCharsets.UTF_8)).toString();
    }

    @Override
    public Map<String, String> retrieveState(final String instanceId, final StateScope scope) {
        final String documentId = getDocumentId(instanceId, scope);
        try {
            final CosmosItemResponse<StateDocument> itemResponse = container.readItem(documentId, new PartitionKey(documentId), StateDocument.class);
            if (itemResponse.getItem() == null) {
                logger.info("State has not yet been stored for {}, scope {}", instanceId, scope);
                return null;
            }

            return itemResponse.getItem().getState();
        } catch (final NotFoundException e) {
            logger.info("State has not been stored for {}, scope {}", instanceId, scope);
            return null;
        }
    }

    public static class StateDocument {
        private String id;
        private String scope;
        private Map<String, String> state;

        public StateDocument() {

        }

        public StateDocument(final String documentId, final String scope, final Map<String, String> state) {
            this.id = documentId;
            this.scope = scope;
            this.state = state;
        }

        public String getId() {
            return id;
        }

        public String getScope() {
            return scope;
        }

        public Map<String, String> getState() {
            return state;
        }

        public void setId(String id) {
            this.id = id;
        }

        public void setScope(String scope) {
            this.scope = scope;
        }

        public void setState(Map<String, String> state) {
            this.state = state;
        }
    }
}
