package com.cloudera.naaf;

import com.cloudera.naaf.metering.CloudPlatform;

import java.nio.charset.StandardCharsets;
import java.util.UUID;

/**
 * Generates a deterministic function ID using a name-based UUID from
 * the string "flowCRN+functionName+cloudProvider".
 */
public class StandardFunctionIdProvider implements FunctionIdProvider {

    private static final String VERSION_SEPARATOR = "/v.";

    /**
     * Generates a deterministic function ID using a name-based UUID from
     * the string "flowCRN+functionName+cloudProvider".
     *
     * @return a UUID based on "flowCRN+functionName+cloudProvider"
     */
    @Override
    public String generateFunctionId(String flowCrn, String functionName, CloudPlatform cloudPlatform) {

        if (flowCrn == null || functionName == null || cloudPlatform == null) {
            throw new IllegalArgumentException("The following fields are required: flowCrn, functionName, cloudPlatform");
        }

        // If the flowCrn is a actually a flowVersionCrn, drop the version suffix,
        // as the functionCrn should not change based on flow version.
        final int lastIndexOfVersionSeparator = flowCrn.lastIndexOf(VERSION_SEPARATOR);
        final boolean isFlowVersionCrn = lastIndexOfVersionSeparator >= 0;
        final String normalizedFlowCrn = isFlowVersionCrn
                ? flowCrn.substring(0, lastIndexOfVersionSeparator).trim().toLowerCase()
                : flowCrn.trim().toLowerCase();
        final String normalizedFunctionName = functionName.trim().toLowerCase();
        final String normalizedCloudProvider = cloudPlatform.getCloudProvider();

        if (normalizedFlowCrn.isEmpty() || normalizedFunctionName.isEmpty() || normalizedCloudProvider.isEmpty()) {
            throw new IllegalArgumentException("The following fields cannot be blank: flowCrn, functionName, cloudPlatform");
        }

        final String functionIdInputString = String.format("%s+%s+%s", normalizedFlowCrn, normalizedFunctionName, normalizedCloudProvider);
        final String functionId = UUID.nameUUIDFromBytes(functionIdInputString.getBytes(StandardCharsets.UTF_8)).toString();
        return functionId;
    }

}
