/*
 * (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf;

import com.cloudera.naaf.metering.MeteringException;
import org.apache.nifi.flowfile.FlowFile;
import org.apache.nifi.stateless.flow.TriggerResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLClassLoader;
import java.util.Arrays;

public class StatelessNiFiUtil {
    private static final Logger logger = LoggerFactory.getLogger(StatelessNiFiUtil.class);

    private StatelessNiFiUtil() {

    }

    /**
     * Handles the trigger result, either acknowledging the result if successful, or throwing an exception.
     * @param triggerResult A dataflow trigger result
     */
    public static void handleTriggerResult(final TriggerResult triggerResult) {
        if (triggerResult.isSuccessful()) {
            logger.debug("Dataflow completed successfully");
            triggerResult.acknowledge();
            logger.debug("Acknowledged trigger result{}", triggerResult.getFailureCause().isPresent() ? " despite failure" : "");
        }

        final Throwable failure = !triggerResult.isSuccessful() ? new RuntimeException("Failed to execute dataflow")
                : triggerResult.getFailureCause().orElse(null);
        // This can happen with a successful invocation if the metering publication failed, but we still need to acknowledge
        // the successful invocation or the next invocation will hang
        if (failure != null) {
            final RuntimeException exception = (failure instanceof RuntimeException)
                    ? (RuntimeException) failure : new RuntimeException("Failed to execute dataflow", failure);
            if (failure instanceof MeteringException) {
                logger.error(failure.getMessage());
            } else {
                logger.error("Failure in the dataflow", exception);
            }

            throw exception;
        }
    }

    /**
     * Copies the output from the output flow file, if it exists, to the provided output stream.
     * @param outputFlowfile A flow file whose contents should be copied
     * @param triggerResult The dataflow trigger result
     * @param outputStream An output stream to which to write
     * @throws IOException In case of an I/O exception
     */
    public static void copyOutput(final FlowFile outputFlowfile, final TriggerResult triggerResult,
                                  final OutputStream outputStream) throws IOException {

        if (outputStream == null) {
            throw new IllegalArgumentException("Output stream must be provided");
        }

        if (outputFlowfile == null) {
            logger.debug("No Output FlowFile so will not write any content to the output stream");
        } else {
            try (final InputStream flowFileContent = triggerResult.readContent(outputFlowfile)) {
                final long numBytes = copy(flowFileContent, outputStream);
                logger.debug("Wrote {} bytes to the output stream from the Output FlowFile's contents", numBytes);
            }
        }
    }

    public static void logClassLoader(final String message, final ClassLoader classLoader) {
        if (classLoader instanceof URLClassLoader) {
            final URLClassLoader urlClassLoader = (URLClassLoader) classLoader;
            logger.debug("ClassLoader {} is {} (with parent {}) has the following URLs available: {}",
                    message, urlClassLoader, urlClassLoader.getParent(), Arrays.asList(urlClassLoader.getURLs()));
        } else {
            logger.debug("The {} ClassLoader {} is {} (with parent {})", message, classLoader, classLoader.getParent());
        }
    }

    private static long copy(final InputStream in, final OutputStream out) throws IOException {
        long totalBytes = 0L;

        int len;
        final byte[] buffer = new byte[4096];
        while ((len = in.read(buffer)) > 0) {
            out.write(buffer, 0, len);
            totalBytes += len;
        }

        return totalBytes;
    }
}
