package metering;

import com.cloudera.naaf.FunctionIdProvider;
import com.cloudera.naaf.StandardFunctionIdProvider;
import org.junit.Test;

import static com.cloudera.naaf.metering.CloudPlatform.AWS_LAMBDA;
import static com.cloudera.naaf.metering.CloudPlatform.AZURE_FUNCTIONS;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

public class StandardFunctionIdProviderTest {

    private static final String FLOW1_CRN = "crn:cdp:df:us-west-1:9d74eee4-1cad-45d7-b645-7ccf9edbb73d:flow:Kevin-Inbound-MiNiFi";
    private static final String FLOW1_V1_CRN = "crn:cdp:df:us-west-1:9d74eee4-1cad-45d7-b645-7ccf9edbb73d:flow:Kevin-Inbound-MiNiFi/v.1";
    private static final String FLOW1_V2_CRN = "crn:cdp:df:us-west-1:9d74eee4-1cad-45d7-b645-7ccf9edbb73d:flow:Kevin-Inbound-MiNiFi/v.2";
    private static final String FLOW1_FUNCTION1_NAME = "Kevin Inbound MiNiFi Function 1";
    private static final String FLOW1_FUNCTION2_NAME = "Kevin Inbound MiNiFi Function 2";
    private static final String FLOW2_CRN = "crn:cdp:df:us-west-1:9d74eee4-1cad-45d7-b645-7ccf9edbb73d:flow:Inbound-UDP-ListenFlow";

    private final FunctionIdProvider functionIdProvider = new StandardFunctionIdProvider();

    @Test(expected = IllegalArgumentException.class)
    public void testNullInputValidation() {
        functionIdProvider.generateFunctionId(null, null, null);
    }

    @Test(expected = IllegalArgumentException.class)
    public void testBlankInputValidation() {
        functionIdProvider.generateFunctionId(" ", " ", AWS_LAMBDA);
    }

    @Test
    public void testMatchingInputsResultInMatchingIds() {
        assertEquals(
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA));
    }

    @Test
    public void testLeadingOrTrailingWhitespaceDoesNotChangeId() {
        assertEquals(
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(
                        " \t\n" + FLOW1_CRN + " \t\n",
                        " \t\n" + FLOW1_FUNCTION1_NAME + " \t\n",
                        AWS_LAMBDA));
    }

    @Test
    public void testCaseDoesNotChangeId() {
        assertEquals(
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(FLOW1_CRN.toUpperCase(), FLOW1_FUNCTION1_NAME.toUpperCase(), AWS_LAMBDA));
    }

    @Test
    public void testFlowCrnChangesFunctionId() {
        assertNotEquals(
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(FLOW2_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA));
    }

    @Test
    public void testFlowVersionCrnDoesNotChangeFunctionId() {
        assertEquals(
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(FLOW1_V1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA));

        assertEquals(
                functionIdProvider.generateFunctionId(FLOW1_V1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(FLOW1_V2_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA));
    }

    @Test
    public void testFunctionNameChangesFunctionId() {
        assertNotEquals(
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION2_NAME, AWS_LAMBDA));
    }

    @Test
    public void testCloudProviderChangesFunctionId() {
        assertNotEquals(
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AWS_LAMBDA),
                functionIdProvider.generateFunctionId(FLOW1_CRN, FLOW1_FUNCTION1_NAME, AZURE_FUNCTIONS));
    }

}