/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.gcp.cloud.functions;

import com.cloudera.naaf.StatelessNiFiFunction;
import com.cloudera.naaf.StatelessNiFiUtil;
import com.google.cloud.functions.HttpFunction;
import com.google.cloud.functions.HttpRequest;
import com.google.cloud.functions.HttpResponse;
import org.apache.nifi.flowfile.FlowFile;
import org.apache.nifi.stateless.flow.TriggerResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.concurrent.atomic.AtomicReference;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class StatelessNiFiHttpFunction implements HttpFunction {
    private static final Logger logger = LoggerFactory.getLogger(StatelessNiFiHttpFunction.class);


    // Environment variables
    private static final String HEADER_ATTRIBUTE_PATTERN_ENV_VAR = "HEADER_ATTRIBUTE_PATTERN";
    private static final String HTTP_STATUS_CODE_ATTRIBUTE_ENV_VAR = "HTTP_STATUS_CODE_ATTRIBUTE";

    private static final String MIME_TYPE = "mime.type";
    private static final String TEXT_PLAIN = "text/plain";
    private static final String RESPONSE_IO_ERROR_MSG = "Could not write HTTP Response";
    private static final String ATTRIBUTE_PREFIX = "gcp.http.header.";

    private static final int STATUS_OK = 200;
    private static final int STATUS_INTERNAL_SERVER_ERROR = 500;

    private static final AtomicReference<StatelessNiFiGcpFunctionHolder> functionHolderReference = new AtomicReference<>();
    private final Pattern headerAttributePattern;

    public StatelessNiFiHttpFunction() {
        final String headerAttributePatternVariable = System.getenv(HEADER_ATTRIBUTE_PATTERN_ENV_VAR);
        headerAttributePattern = headerAttributePatternVariable == null ? null : Pattern.compile(headerAttributePatternVariable);
    }

    @Override
    public void service(final HttpRequest httpRequest, final HttpResponse httpResponse) throws IOException {
        final StatelessNiFiGcpFunctionHolder functionHolder = getFunctionHolder();
        final StatelessNiFiFunction statelessNiFiFunction = functionHolder.getStatelessNiFiFunction();

        try {
            logger.debug("Triggering HTTP Function");
            final String instanceId = functionHolder.getFunctionName();
            final TriggerResult triggerResult = statelessNiFiFunction.trigger(instanceId, httpRequest.getInputStream(), getAttributes(httpRequest));

            logger.debug("Triggered HTTP Function, successful = {}", triggerResult.isSuccessful());
            if (triggerResult.isSuccessful()) {
                final FlowFile outputFlowFile = statelessNiFiFunction.getOutputFlowfile(triggerResult);
                final Map<String, String> attributes = outputFlowFile == null ? null : outputFlowFile.getAttributes();

                if (attributes == null) {
                    writeDefaultHttpResponse(httpResponse);
                } else {
                    prepareHttpResponse(httpResponse, attributes);
                    StatelessNiFiUtil.copyOutput(outputFlowFile, triggerResult, httpResponse.getOutputStream());
                }

                triggerResult.acknowledge();
                logger.debug("Acknowledged Trigger Result");
            } else {
                logger.error("Trigger dataflow but received a failure result", triggerResult.getFailureCause().orElse(null));
                throw triggerResult.getFailureCause().orElse(new RuntimeException("DataFlow failed"));
            }
        } catch (final Throwable t) {
            logger.error("Failed to execute dataflow", t);
            writeErrorHttpResponse(httpResponse, t);
        }
    }

    private StatelessNiFiGcpFunctionHolder getFunctionHolder() throws IOException {
        StatelessNiFiGcpFunctionHolder functionHolder = functionHolderReference.get();
        if (functionHolder == null) {
            logger.debug("Function has not been initialized yet, initializing now...");
            functionHolder = new StatelessNiFiGcpFunctionHolder();
            functionHolderReference.set(functionHolder);
            return functionHolder;
        }

        logger.debug("Function already initialized.");
        return functionHolder;
    }

    private Map<String, String> getAttributes(final HttpRequest httpRequest) {
        final Map<String, String> attributes = httpRequest.getHeaders().entrySet().stream()
                .filter(entry -> httpRequest.getFirstHeader(entry.getKey()).isPresent())
                .collect(Collectors.toMap(
                        entry -> ATTRIBUTE_PREFIX + entry.getKey(),
                        entry -> httpRequest.getFirstHeader(entry.getKey()).get()
                ));

        if (httpRequest.getContentType().isPresent()) {
            attributes.put(MIME_TYPE, httpRequest.getContentType().get());
        }
        attributes.put(StatelessNiFiFunction.CLOUD_PROVIDER_ATTRIBUTE, StatelessNiFiGcpFunctionHolder.GCP_CLOUD_PROVIDER);
        attributes.put("gcp.http.method", httpRequest.getMethod());

        return attributes;
    }

    private void prepareHttpResponse(final HttpResponse httpResponse, final Map<String, String> outputAttributes) {
        int statusCode = STATUS_OK;
        if (outputAttributes.containsKey(MIME_TYPE)) {
            httpResponse.setContentType(outputAttributes.get(MIME_TYPE));
        }

        final String httpStatusCodeAttribute = System.getenv(HTTP_STATUS_CODE_ATTRIBUTE_ENV_VAR);
        if (httpStatusCodeAttribute != null) {
            final String statusCodeValue = outputAttributes.get(httpStatusCodeAttribute);
            logger.debug("HTTP Status Code Value = {}", statusCodeValue);

            if (statusCodeValue != null) {
                try {
                    final int parsedStatusCode = Integer.parseInt(statusCodeValue);
                    if (parsedStatusCode < 100 || parsedStatusCode >= 600) {
                        final String errorMessage = String.format("Invalid HTTP status code [%s] provided in [%s] attribute", parsedStatusCode, httpStatusCodeAttribute);
                        logger.error(errorMessage);
                        throw new IllegalArgumentException(errorMessage);
                    }

                    statusCode = parsedStatusCode;
                } catch (final NumberFormatException e) {
                    final String errorMessage = String.format("Could not parse HTTP status code attribute [%s] value", httpStatusCodeAttribute);
                    logger.error(errorMessage, e);
                    throw new RuntimeException(errorMessage, e);
                }
            }
        }

        if (headerAttributePattern != null) {
            outputAttributes.entrySet().stream()
                .filter(entry -> headerAttributePattern.matcher(entry.getKey()).matches())
                .forEach(entry -> httpResponse.appendHeader(entry.getKey(), entry.getValue()));
        }

        httpResponse.setStatusCode(statusCode);
    }

    private void writeDefaultHttpResponse(final HttpResponse httpResponse) {
        try (final OutputStream output = httpResponse.getOutputStream()) {
            httpResponse.setStatusCode(STATUS_OK);
            httpResponse.setContentType(TEXT_PLAIN);
            output.write("DataFlow completed successfully".getBytes(StandardCharsets.UTF_8));
        } catch (final IOException e) {
            logger.error(RESPONSE_IO_ERROR_MSG, e);
            throw new RuntimeException(RESPONSE_IO_ERROR_MSG, e);
        }
    }

    private void writeErrorHttpResponse(final HttpResponse httpResponse, final Throwable t) {
        try (final OutputStream output = httpResponse.getOutputStream()) {
            httpResponse.setStatusCode(STATUS_INTERNAL_SERVER_ERROR);
            httpResponse.setContentType(TEXT_PLAIN);
            output.write(("DataFlow failure: " + t.getMessage()).getBytes(StandardCharsets.UTF_8));
        } catch (final IOException e) {
            logger.error(RESPONSE_IO_ERROR_MSG, e);
            throw new RuntimeException(RESPONSE_IO_ERROR_MSG, e);
        }
    }
}
