/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.gcp.cloud.functions;

import com.cloudera.naaf.StatelessNiFiStateProvider;
import com.google.api.core.ApiFuture;
import com.google.auth.oauth2.GoogleCredentials;
import com.google.cloud.firestore.DocumentReference;
import com.google.cloud.firestore.DocumentSnapshot;
import com.google.cloud.firestore.Firestore;
import com.google.cloud.firestore.FirestoreOptions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ExecutionException;

public class FirestoreStatelessNiFiStateProvider implements StatelessNiFiStateProvider {
    private static final Logger logger = LoggerFactory.getLogger(FirestoreStatelessNiFiStateProvider.class);

    public static final String DEFAULT_STATE_COLLECTION = "nifi_state";
    public static final String INSTANCE_ID = "instanceId";
    public static final String SCOPE = "scope";
    public static final String STATE = "state";

    public static final String FIRESTORE_STATE_COLLECTION = "FIRESTORE_STATE_COLLECTION";
    public static final String PROJECT_ID_VAR = "GCP_PROJECT";

    private final String collection;

    private Firestore firestore;

    public FirestoreStatelessNiFiStateProvider() {
        final String projectId = System.getenv(PROJECT_ID_VAR);
        collection = Optional.ofNullable(System.getenv(FIRESTORE_STATE_COLLECTION)).orElse(DEFAULT_STATE_COLLECTION);

        try {
            final FirestoreOptions firestoreOptions =
                    FirestoreOptions.getDefaultInstance().toBuilder()
                            .setProjectId(projectId)
                            .setCredentials(GoogleCredentials.getApplicationDefault())
                            .build();
            firestore = firestoreOptions.getService();
        } catch (final IOException e) {
            logger.error("Could not create Firestore Client", e);
            throw new RuntimeException("Could not create Firestore Client", e);
        }
    }

    @Override
    public void storeState(final String instanceId, final StateScope scope, final Map<String, String> state) {
        final DocumentReference document = firestore.collection(collection).document(getDocumentId(instanceId, scope));
        document.set(state);
    }

    @Override
    public Map<String, String> retrieveState(final String instanceId, final StateScope scope) {
        final String documentId = getDocumentId(instanceId, scope);
        final DocumentReference document = firestore.collection(collection).document(documentId);
        final ApiFuture<DocumentSnapshot> documentSnapshot = document.get();
        try {
            final Map<String, Object> state = documentSnapshot.get().getData();
            if (state == null) {
                logger.info("State attribute has not been stored for {}, scope {}", instanceId, scope);
                return null;
            }
            final Map<String, String> stateMap = new HashMap<>();
            state.forEach((key, value) -> stateMap.put(key, value.toString()));
            return stateMap;
        } catch (final InterruptedException | ExecutionException e) {
            logger.error("Could not retrieve state from Firestore for document {}", documentId);
            throw new RuntimeException("Could not retrieve state from Firestore for document " + documentId);
        }
    }

    private String getDocumentId(final String instanceId, final StateScope scope) {
        return instanceId + "-" + scope;
    }
}
