/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.gcp.cloud.functions.resources;

import com.cloudera.naaf.resources.AbstractResourceDownloader;
import com.google.api.gax.paging.Page;
import com.google.cloud.storage.Blob;
import com.google.cloud.storage.Storage;
import com.google.cloud.storage.StorageOptions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.file.Path;

public class GcpCloudStorageResourceDownloader extends AbstractResourceDownloader {
    private static final Logger logger = LoggerFactory.getLogger(GcpCloudStorageResourceDownloader.class);

    private static final String STORAGE_BUCKET_ENV_VAR = "STORAGE_BUCKET";

    @Override
    public void download(final String sourceDirectory, final File downloadDirectory) {
        final String bucketName = System.getenv(STORAGE_BUCKET_ENV_VAR);
        final String bucketPrefix = sourceDirectory.endsWith("/") ? sourceDirectory : sourceDirectory + "/";

        final Storage storage = StorageOptions.getDefaultInstance().getService();

        Page<Blob> page = storage.list(bucketName, Storage.BlobListOption.prefix(bucketPrefix));

        int downloadCount = 0;
        do {
            for (final Blob blob : page.getValues()) {
                final String blobName = blob.getName();
                final String filenameWithinSourceDir = substringAfter(blobName, bucketPrefix);
                final Path filePath = getDownloadPath(filenameWithinSourceDir, downloadDirectory);
                if (filePath == null) {
                    continue;
                }

                logger.info("Downloading {} to {}", blobName, filePath);
                blob.downloadTo(filePath);
                downloadCount++;
            }
        } while ((page = page.getNextPage()) != null);

        if (downloadCount == 0) {
            logger.debug("Found no blobs in {}/{}", bucketName, bucketPrefix);
        }
    }
}
