/*
 * (c) 2018-2022 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.metering;

import com.cloudera.cdp.authentication.credentials.BasicCdpCredentials;
import com.cloudera.cdp.authentication.credentials.CdpCredentials;
import com.cloudera.cdp.authentication.credentials.StaticCdpCredentialsProvider;
import com.cloudera.cdp.dbus.api.DbusClient;
import com.cloudera.cdp.dbus.api.DbusClientBuilder;
import com.cloudera.cdp.dbus.model.Header;
import com.cloudera.cdp.dbus.model.PutRecordRequest;
import com.cloudera.cdp.dbus.model.Record;
import com.cloudera.cdp.shaded.com.fasterxml.jackson.annotation.JsonInclude;
import com.cloudera.cdp.shaded.com.fasterxml.jackson.core.JsonProcessingException;
import com.cloudera.cdp.shaded.com.fasterxml.jackson.databind.ObjectMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;
import java.util.concurrent.TimeUnit;

public class DbusMeteringPublisher implements MeteringPublisher {
    private enum DbusCluster {
        PRIV("manowar_dev", "mow-priv", "https://dbusapi.sigma-dev.cloudera.com:443"),  // priv uses dev env and dev dbus api with different stream name
        DEV("manowar_dev", "mow", "https://dbusapi.sigma-dev.cloudera.com:443"),
        INT("manowar_int", "mow-int", "https://dbusapi.sigma-int.cloudera.com:443"),
        STAGE("manowar_stage", "mow-stage", "https://dbusapi.sigma-stage.cloudera.com:443"),
        PROD("manowar_prod", "mow-prod", "https://dbusapi.us-west-1.sigma.altus.cloudera.com");
        // TODO we will have to update PROD dbusEndpoint to be based on region for multi-region control plane support

        private String environment;
        private String cluster;
        private String dbusEndpoint;

        DbusCluster(final String environment, final String cluster, final String dbusEndpoint) {
            this.environment = environment;
            this.cluster = cluster;
            this.dbusEndpoint = dbusEndpoint;
        }
    }

    private static final Logger logger = LoggerFactory.getLogger(DbusMeteringPublisher.class);
    private static final int VERSION = 1;

    private static final String STREAM_NAME = "NaafMeter";
    private static final String DEFAULT_STREAM_APP_NAME = "NaafMeter";
    private static final String QUALIFIED_STREAM_NAME_FORMAT = "%s-%s-%s";
    private static final String ACCESS_KEY_ENV_VAR = "DF_ACCESS_KEY";
    private static final String PRIVATE_KEY_ENV_VAR = "DF_PRIVATE_KEY";
    private static final String STREAM_APP_NAME_ENV_VAR = "NAAF_STREAM_APP";

    // This can be set to dev, int, or stage for internal DBus testing, but defaults to the production DBus endpoint
    private static final String DBUS_CLUSTER_ENV_VAR = "DBUS_CLUSTER";

    private static final String APPLICATION_HEADER_NAME = "Applications";
    private static final String APPLICATION_HEADER_VALUE = "DataFlow Functions";
    private static final String STREAM_APP_HEADER_NAME = "@naafmeter-app";

    private final ObjectMapper objectMapper = new ObjectMapper();
    private DbusClient dbusClient;
    private String accessKeyId;
    private String streamName;
    private String streamAppName;

    @Override
    public void initialize() {
        this.accessKeyId = System.getenv(ACCESS_KEY_ENV_VAR);
        final String privateKey = System.getenv(PRIVATE_KEY_ENV_VAR);
        final String streamAppName = System.getenv(STREAM_APP_NAME_ENV_VAR);
        final String dbusClusterEnvVarValue = System.getenv(DBUS_CLUSTER_ENV_VAR);

        final DbusCluster dbusCluster = dbusClusterEnvVarValue != null
                ? DbusCluster.valueOf(dbusClusterEnvVarValue.toUpperCase())
                : DbusCluster.PROD;

        // For unique streamAppName, e.g., in mow-priv/moonlander private stacks
        this.streamAppName = streamAppName != null
                ? getQualifiedStreamAppName(dbusCluster.environment, dbusCluster.cluster, streamAppName)
                : getQualifiedStreamAppName(dbusCluster.environment, dbusCluster.cluster, DEFAULT_STREAM_APP_NAME);

        this.streamName = getQualifiedStreamName(dbusCluster.environment, dbusCluster.cluster);

        final CdpCredentials credentials = new BasicCdpCredentials(accessKeyId, privateKey);

        this.dbusClient = DbusClientBuilder.defaultBuilder()
                .withCredentials(new StaticCdpCredentialsProvider(credentials))
                .withEndPoint(dbusCluster.dbusEndpoint)
                .build();

        this.objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
    }

    @Override
    public synchronized void publish(final Invocation invocation) {
        try {
            DbusNaafInvocation dbusNaafInvocation = getNaafInvocation(invocation);

            final PutRecordRequest putRecordRequest =
                    createRequest(getPayload(dbusNaafInvocation), dbusNaafInvocation.getFunctionId());

            logger.debug("Submitting to metering service: " + putRecordRequest);
            dbusClient.putRecord(putRecordRequest);
            logger.debug("Successfully submitted billing invocation with partition key {} to metering service", putRecordRequest.getRecord().getPartitionKey());
        } catch (final Throwable e) {
            logger.error("Failed to submit billing invocation to metering service", e);
            throw new MeteringException("Failed to submit billing invocation to metering service", e, invocation.isSuccessful());
        }
    }

    private DbusNaafInvocation getNaafInvocation(final Invocation invocation) {
        final DbusNaafInvocation dbusNaafInvocation = new DbusNaafInvocation();
        dbusNaafInvocation.setDurationMillis(invocation.getDuration(TimeUnit.MILLISECONDS));
        dbusNaafInvocation.setFlowCrn(invocation.getFlowCrn());
        dbusNaafInvocation.setFunctionId(invocation.getFunctionId());
        dbusNaafInvocation.setFunctionName(invocation.getFunctionName());
        dbusNaafInvocation.setInvocationId(invocation.getId());
        dbusNaafInvocation.setPlatform(invocation.getCloudPlatform());
        dbusNaafInvocation.setStartTime(invocation.getStartTime().toEpochMilli());
        dbusNaafInvocation.setSuccess(invocation.isSuccessful());
        dbusNaafInvocation.setVersion(VERSION);
        if (invocation.getRegion() == null) {
            dbusNaafInvocation.setRegion("UNKNOWN");
        } else {
            dbusNaafInvocation.setRegion(invocation.getRegion());
        }
        dbusNaafInvocation.setAccessKeyId(accessKeyId);
        return dbusNaafInvocation;
    }

    private String getPayload(final DbusNaafInvocation dbusNaafInvocation) {
        try {
            return objectMapper.writeValueAsString(dbusNaafInvocation);
        } catch (final JsonProcessingException e) {
            throw new RuntimeException("Failed to serialize billed invocation", e);
        }
    }

    private PutRecordRequest createRequest(String payload, String functionId) {

        final List<Header> headers = new ArrayList<>();
        headers.add(getHeader(APPLICATION_HEADER_NAME, APPLICATION_HEADER_VALUE));
        headers.add(getHeader(STREAM_APP_HEADER_NAME, streamAppName));

        final PutRecordRequest putRecordRequest = new PutRecordRequest();
        final Record record = new Record();
        record.setPayload(Base64.getEncoder().encodeToString(payload.getBytes(StandardCharsets.UTF_8)));
        record.setPayloadSize(Long.valueOf(payload.length()));
        record.setPartitionKey(functionId);
        record.setStreamName(streamName);
        record.setHeaders(headers);
        putRecordRequest.setRecord(record);
        return putRecordRequest;
    }

    private Header getHeader(String name, String value) {
        Header header = new Header();
        header.setName(name);
        header.setValue(value);
        return header;
    }

    private String getQualifiedStreamName(final String environment, final String cluster) {
        return String.format(QUALIFIED_STREAM_NAME_FORMAT, environment, cluster, STREAM_NAME);
    }

    private String getQualifiedStreamAppName(final String environment, final String cluster, final String streamAppName) {
        return String.format(QUALIFIED_STREAM_NAME_FORMAT, environment, cluster, streamAppName);
    }
}
