package com.cloudera.naaf.metering;

import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

/**
 * Before running this test, add the following env variables to this test's configuration in your IDE:
 * FLOW_CRN={flowCrn}
 *   dev:   crn:cdp:df:us-west-1:9d74eee4-1cad-45d7-b645-7ccf9edbb73d:flow:testflow
 *   int:   crn:cdp:df:us-west-1:c8dbde4b-ccce-4f8d-a581-830970ba4908:flow:simple
 *   stage:
 *   prod:  crn:cdp:df:us-west-1:558bc1d2-8867-4357-8524-311d51259233:flow:a-sample-flow
 * DF_ACCESS_KEY={accessKey}
 * DF_PRIVATE_KEY={privateKey}
 * DBUS_CLUSTER=priv|dev|int|stage|prod
 **/

@Ignore("For manual testing purposes only.")
public class DbusMeteringPublisherTest {

    private final DbusMeteringPublisher dbusBillingPublisher = new DbusMeteringPublisher();
    private String flowCrn;

    @Before
    public void setUp() {
        flowCrn = System.getenv("FLOW_CRN");
        dbusBillingPublisher.initialize();
    }

    @Test
    public void testPublishWithRegion() {

        String functionId = "6c34eee4-1cad-45d7-b645-7ccf9edbb747";

        // Arrange
        Invocation invocation = StandardInvocation.builder()
                .startTime(Instant.now())
                .durationNanos(TimeUnit.SECONDS.toNanos(1))
                .functionName("Test_function")
                .functionId(functionId)
                .region("us-west2")
                .cloudPlatform(CloudPlatform.GOOGLE_CLOUD_FUNCTIONS)
                .flowCrn(flowCrn)
                .successful(true)
                .build();

        // Act
        dbusBillingPublisher.publish(invocation);

    }

    // This code is designed for generating a lot of test invocation records in
    // mow-dev over a long period of time for integration testing. It runs forever
    // until stopped. To enable it in your IDE, uncomment the @Test annotation.
    //@Test
    public void endlessTest() throws Exception {

        // Generate functions
        DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
        List<StandardInvocation.Builder> functions = new ArrayList<>(100);
        List<Stats> expectedStats = new ArrayList<>(100);

        for (int i = 1; i <= 100; i++) {

            String functionName = "Function-" + i;
            String functionId = UUID.randomUUID().toString();
            StandardInvocation.Builder builder = StandardInvocation.builder()
                    .functionName(functionName)
                    .functionId(functionId)
                    .flowCrn(flowCrn)
                    .region("UNKNOWN")
                    .durationNanos(TimeUnit.SECONDS.toNanos(i)) // duration will be 'i' seconds
                    .successful(true);


            CloudPlatform cloudPlatform;
            int cloudCodeSelector = i % 3;
            if (cloudCodeSelector == 0) {
                cloudPlatform = CloudPlatform.AZURE_FUNCTIONS;
            } else if (cloudCodeSelector == 1) {
                cloudPlatform = CloudPlatform.GOOGLE_CLOUD_FUNCTIONS;
            } else {
                cloudPlatform = CloudPlatform.AWS_LAMBDA;
            }
            builder.cloudPlatform(cloudPlatform);

            functions.add(builder);
            expectedStats.add(new Stats(functionName, functionId));
        }

        // Every minute, publish a new set of invocations
        while (true) {
            System.out.println("*********************************************************************************************");
            System.out.println("******************* SENDING NEW BATCH OF INVOCATIONS - DO NOT STOP TEST *********************");
            System.out.println("*********************************************************************************************");
            System.out.println("Now: "+ dtf.format(LocalDateTime.now()));
            System.out.println("Now (epoch millis): " + System.currentTimeMillis());
            System.out.println();
            for (int i = 0; i <= 99; i++) {
                Invocation invocation = functions.get(i).startTime(Instant.now()).build();
                System.out.println("Sending Function-" + i + " [" + invocation.getFunctionId() + "]: duration=" + invocation.getDuration(TimeUnit.MILLISECONDS) + "ms");
                dbusBillingPublisher.publish(invocation);
                expectedStats.get(i).add(invocation);
                Thread.sleep(5);
            }

            // Print aggregate duration and expected billed invocations per function\
            System.out.println("\n******************** EXPECTED RESULTS BASED ON TOTAL DATA SENT SO FAR ***********************");
            for (Stats s : expectedStats) {
                System.out.println(s);
            }
            System.out.println("\nNow: "+ dtf.format(LocalDateTime.now()));
            System.out.println("Now (epoch millis): " + System.currentTimeMillis());
            System.out.println("\nSleeping for 30 seconds, during which it is safe to kill this test.\n");
            Thread.sleep(TimeUnit.SECONDS.toMillis(30));
        }
    }

    private static class Stats {
        private final String functionName;
        private final String functionId;
        private long invocations;
        private long billedInvocations;

        public Stats(String functionName, String functionId) {
            this.functionName = functionName;
            this.functionId = functionId;
            this.invocations = 0;
            this.billedInvocations = 0;
        }

        public void add(Invocation invocation) {
            this.invocations++;
            this.billedInvocations += Math.max(1, invocation.getDuration(TimeUnit.SECONDS));
        }

        @Override
        public String toString() {
            return functionName + " [" + functionId + "]" + ": " +
                    "invocations=" + invocations +
                    ", billedInvocations=" + billedInvocations;
        }
    }

}
